#include "bifrost_nir.h"

#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 29
 * transforms:
 *    ('ineg', 'a') => ('isub', 0, 'a')
 *    ('b2f16', 'a@8') => ('b8csel', 'a', 1.0, 0.0)
 *    ('b2f32', 'a@8') => ('b8csel', 'a', 1.0, 0.0)
 *    ('b2f64', 'a@8') => ('b8csel', 'a', 1.0, 0.0)
 *    ('b2f16', 'a@16') => ('b16csel', 'a', 1.0, 0.0)
 *    ('b2f32', 'a@16') => ('b16csel', 'a', 1.0, 0.0)
 *    ('b2f64', 'a@16') => ('b16csel', 'a', 1.0, 0.0)
 *    ('b2f16', 'a@32') => ('b32csel', 'a', 1.0, 0.0)
 *    ('b2f32', 'a@32') => ('b32csel', 'a', 1.0, 0.0)
 *    ('b2f64', 'a@32') => ('b32csel', 'a', 1.0, 0.0)
 *    ('imin', 'a@8', 'b@8') => ('b8csel', ('ilt8', 'a', 'b'), 'a', 'b')
 *    ('imax', 'a@8', 'b@8') => ('b8csel', ('ilt8', 'b', 'a'), 'a', 'b')
 *    ('umin', 'a@8', 'b@8') => ('b8csel', ('ult8', 'a', 'b'), 'a', 'b')
 *    ('umax', 'a@8', 'b@8') => ('b8csel', ('ult8', 'b', 'a'), 'a', 'b')
 *    ('imin', 'a@16', 'b@16') => ('b16csel', ('ilt16', 'a', 'b'), 'a', 'b')
 *    ('imax', 'a@16', 'b@16') => ('b16csel', ('ilt16', 'b', 'a'), 'a', 'b')
 *    ('umin', 'a@16', 'b@16') => ('b16csel', ('ult16', 'a', 'b'), 'a', 'b')
 *    ('umax', 'a@16', 'b@16') => ('b16csel', ('ult16', 'b', 'a'), 'a', 'b')
 *    ('imin', 'a@32', 'b@32') => ('b32csel', ('ilt32', 'a', 'b'), 'a', 'b')
 *    ('imax', 'a@32', 'b@32') => ('b32csel', ('ilt32', 'b', 'a'), 'a', 'b')
 *    ('umin', 'a@32', 'b@32') => ('b32csel', ('ult32', 'a', 'b'), 'a', 'b')
 *    ('umax', 'a@32', 'b@32') => ('b32csel', ('ult32', 'b', 'a'), 'a', 'b')
 *    ('fexp2@16', 'a') => ('f2f16', ('fexp2', ('f2f32', 'a')))
 *    ('flog2@16', 'a') => ('f2f16', ('flog2', ('f2f32', 'a')))
 *    ('fsin@16', 'a') => ('f2f16', ('fsin', ('f2f32', 'a')))
 *    ('fcos@16', 'a') => ('f2f16', ('fcos', ('f2f32', 'a')))
 *    ('f2b32', 'a') => ('fneu32', 'a', 0.0)
 *    ('i2b32', 'a') => ('ine32', 'a', 0)
 *    ('b2i32', 'a') => ('iand', 'a@32', 1)
 */


   static const nir_search_variable search0_0 = {
   { nir_search_value_variable, -1 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
};
static const nir_search_expression search0 = {
   { nir_search_value_expression, -1 },
   false, false,
   -1, 0,
   nir_op_ineg,
   { &search0_0.value },
   NULL,
};

   static const nir_search_constant replace0_0 = {
   { nir_search_value_constant, -1 },
   nir_type_int, { 0x0 /* 0 */ },
};

/* replace0_1 -> search0_0 in the cache */
static const nir_search_expression replace0 = {
   { nir_search_value_expression, -1 },
   false, false,
   -1, 0,
   nir_op_isub,
   { &replace0_0.value, &search0_0.value },
   NULL,
};

   static const nir_search_variable search1_0 = {
   { nir_search_value_variable, 8 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
};
static const nir_search_expression search1 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b2f16,
   { &search1_0.value },
   NULL,
};

   /* replace1_0 -> search1_0 in the cache */

static const nir_search_constant replace1_1 = {
   { nir_search_value_constant, 16 },
   nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
};

static const nir_search_constant replace1_2 = {
   { nir_search_value_constant, 16 },
   nir_type_float, { 0x0 /* 0.0 */ },
};
static const nir_search_expression replace1 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b8csel,
   { &search1_0.value, &replace1_1.value, &replace1_2.value },
   NULL,
};

   /* search2_0 -> search1_0 in the cache */
static const nir_search_expression search2 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b2f32,
   { &search1_0.value },
   NULL,
};

   /* replace2_0 -> search1_0 in the cache */

static const nir_search_constant replace2_1 = {
   { nir_search_value_constant, 32 },
   nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
};

static const nir_search_constant replace2_2 = {
   { nir_search_value_constant, 32 },
   nir_type_float, { 0x0 /* 0.0 */ },
};
static const nir_search_expression replace2 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b8csel,
   { &search1_0.value, &replace2_1.value, &replace2_2.value },
   NULL,
};

   /* search3_0 -> search1_0 in the cache */
static const nir_search_expression search3 = {
   { nir_search_value_expression, 64 },
   false, false,
   -1, 0,
   nir_op_b2f64,
   { &search1_0.value },
   NULL,
};

   /* replace3_0 -> search1_0 in the cache */

static const nir_search_constant replace3_1 = {
   { nir_search_value_constant, 64 },
   nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
};

static const nir_search_constant replace3_2 = {
   { nir_search_value_constant, 64 },
   nir_type_float, { 0x0 /* 0.0 */ },
};
static const nir_search_expression replace3 = {
   { nir_search_value_expression, 64 },
   false, false,
   -1, 0,
   nir_op_b8csel,
   { &search1_0.value, &replace3_1.value, &replace3_2.value },
   NULL,
};

   static const nir_search_variable search4_0 = {
   { nir_search_value_variable, 16 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
};
static const nir_search_expression search4 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b2f16,
   { &search4_0.value },
   NULL,
};

   /* replace4_0 -> search4_0 in the cache */

/* replace4_1 -> replace1_1 in the cache */

/* replace4_2 -> replace1_2 in the cache */
static const nir_search_expression replace4 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b16csel,
   { &search4_0.value, &replace1_1.value, &replace1_2.value },
   NULL,
};

   /* search5_0 -> search4_0 in the cache */
static const nir_search_expression search5 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b2f32,
   { &search4_0.value },
   NULL,
};

   /* replace5_0 -> search4_0 in the cache */

/* replace5_1 -> replace2_1 in the cache */

/* replace5_2 -> replace2_2 in the cache */
static const nir_search_expression replace5 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b16csel,
   { &search4_0.value, &replace2_1.value, &replace2_2.value },
   NULL,
};

   /* search6_0 -> search4_0 in the cache */
static const nir_search_expression search6 = {
   { nir_search_value_expression, 64 },
   false, false,
   -1, 0,
   nir_op_b2f64,
   { &search4_0.value },
   NULL,
};

   /* replace6_0 -> search4_0 in the cache */

/* replace6_1 -> replace3_1 in the cache */

/* replace6_2 -> replace3_2 in the cache */
static const nir_search_expression replace6 = {
   { nir_search_value_expression, 64 },
   false, false,
   -1, 0,
   nir_op_b16csel,
   { &search4_0.value, &replace3_1.value, &replace3_2.value },
   NULL,
};

   static const nir_search_variable search7_0 = {
   { nir_search_value_variable, 32 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
};
static const nir_search_expression search7 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b2f16,
   { &search7_0.value },
   NULL,
};

   /* replace7_0 -> search7_0 in the cache */

/* replace7_1 -> replace1_1 in the cache */

/* replace7_2 -> replace1_2 in the cache */
static const nir_search_expression replace7 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b32csel,
   { &search7_0.value, &replace1_1.value, &replace1_2.value },
   NULL,
};

   /* search8_0 -> search7_0 in the cache */
static const nir_search_expression search8 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b2f32,
   { &search7_0.value },
   NULL,
};

   /* replace8_0 -> search7_0 in the cache */

/* replace8_1 -> replace2_1 in the cache */

/* replace8_2 -> replace2_2 in the cache */
static const nir_search_expression replace8 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b32csel,
   { &search7_0.value, &replace2_1.value, &replace2_2.value },
   NULL,
};

   /* search9_0 -> search7_0 in the cache */
static const nir_search_expression search9 = {
   { nir_search_value_expression, 64 },
   false, false,
   -1, 0,
   nir_op_b2f64,
   { &search7_0.value },
   NULL,
};

   /* replace9_0 -> search7_0 in the cache */

/* replace9_1 -> replace3_1 in the cache */

/* replace9_2 -> replace3_2 in the cache */
static const nir_search_expression replace9 = {
   { nir_search_value_expression, 64 },
   false, false,
   -1, 0,
   nir_op_b32csel,
   { &search7_0.value, &replace3_1.value, &replace3_2.value },
   NULL,
};

   /* search10_0 -> search1_0 in the cache */

static const nir_search_variable search10_1 = {
   { nir_search_value_variable, 8 },
   1, /* b */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
};
static const nir_search_expression search10 = {
   { nir_search_value_expression, 8 },
   false, false,
   0, 1,
   nir_op_imin,
   { &search1_0.value, &search10_1.value },
   NULL,
};

   /* replace10_0_0 -> search1_0 in the cache */

/* replace10_0_1 -> search10_1 in the cache */
static const nir_search_expression replace10_0 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_ilt8,
   { &search1_0.value, &search10_1.value },
   NULL,
};

/* replace10_1 -> search1_0 in the cache */

/* replace10_2 -> search10_1 in the cache */
static const nir_search_expression replace10 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_b8csel,
   { &replace10_0.value, &search1_0.value, &search10_1.value },
   NULL,
};

   /* search11_0 -> search1_0 in the cache */

/* search11_1 -> search10_1 in the cache */
static const nir_search_expression search11 = {
   { nir_search_value_expression, 8 },
   false, false,
   0, 1,
   nir_op_imax,
   { &search1_0.value, &search10_1.value },
   NULL,
};

   /* replace11_0_0 -> search10_1 in the cache */

/* replace11_0_1 -> search1_0 in the cache */
static const nir_search_expression replace11_0 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_ilt8,
   { &search10_1.value, &search1_0.value },
   NULL,
};

/* replace11_1 -> search1_0 in the cache */

/* replace11_2 -> search10_1 in the cache */
static const nir_search_expression replace11 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_b8csel,
   { &replace11_0.value, &search1_0.value, &search10_1.value },
   NULL,
};

   /* search12_0 -> search1_0 in the cache */

/* search12_1 -> search10_1 in the cache */
static const nir_search_expression search12 = {
   { nir_search_value_expression, 8 },
   false, false,
   0, 1,
   nir_op_umin,
   { &search1_0.value, &search10_1.value },
   NULL,
};

   /* replace12_0_0 -> search1_0 in the cache */

/* replace12_0_1 -> search10_1 in the cache */
static const nir_search_expression replace12_0 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_ult8,
   { &search1_0.value, &search10_1.value },
   NULL,
};

/* replace12_1 -> search1_0 in the cache */

/* replace12_2 -> search10_1 in the cache */
static const nir_search_expression replace12 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_b8csel,
   { &replace12_0.value, &search1_0.value, &search10_1.value },
   NULL,
};

   /* search13_0 -> search1_0 in the cache */

/* search13_1 -> search10_1 in the cache */
static const nir_search_expression search13 = {
   { nir_search_value_expression, 8 },
   false, false,
   0, 1,
   nir_op_umax,
   { &search1_0.value, &search10_1.value },
   NULL,
};

   /* replace13_0_0 -> search10_1 in the cache */

/* replace13_0_1 -> search1_0 in the cache */
static const nir_search_expression replace13_0 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_ult8,
   { &search10_1.value, &search1_0.value },
   NULL,
};

/* replace13_1 -> search1_0 in the cache */

/* replace13_2 -> search10_1 in the cache */
static const nir_search_expression replace13 = {
   { nir_search_value_expression, 8 },
   false, false,
   -1, 0,
   nir_op_b8csel,
   { &replace13_0.value, &search1_0.value, &search10_1.value },
   NULL,
};

   /* search14_0 -> search4_0 in the cache */

static const nir_search_variable search14_1 = {
   { nir_search_value_variable, 16 },
   1, /* b */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
};
static const nir_search_expression search14 = {
   { nir_search_value_expression, 16 },
   false, false,
   0, 1,
   nir_op_imin,
   { &search4_0.value, &search14_1.value },
   NULL,
};

   /* replace14_0_0 -> search4_0 in the cache */

/* replace14_0_1 -> search14_1 in the cache */
static const nir_search_expression replace14_0 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_ilt16,
   { &search4_0.value, &search14_1.value },
   NULL,
};

/* replace14_1 -> search4_0 in the cache */

/* replace14_2 -> search14_1 in the cache */
static const nir_search_expression replace14 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b16csel,
   { &replace14_0.value, &search4_0.value, &search14_1.value },
   NULL,
};

   /* search15_0 -> search4_0 in the cache */

/* search15_1 -> search14_1 in the cache */
static const nir_search_expression search15 = {
   { nir_search_value_expression, 16 },
   false, false,
   0, 1,
   nir_op_imax,
   { &search4_0.value, &search14_1.value },
   NULL,
};

   /* replace15_0_0 -> search14_1 in the cache */

/* replace15_0_1 -> search4_0 in the cache */
static const nir_search_expression replace15_0 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_ilt16,
   { &search14_1.value, &search4_0.value },
   NULL,
};

/* replace15_1 -> search4_0 in the cache */

/* replace15_2 -> search14_1 in the cache */
static const nir_search_expression replace15 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b16csel,
   { &replace15_0.value, &search4_0.value, &search14_1.value },
   NULL,
};

   /* search16_0 -> search4_0 in the cache */

/* search16_1 -> search14_1 in the cache */
static const nir_search_expression search16 = {
   { nir_search_value_expression, 16 },
   false, false,
   0, 1,
   nir_op_umin,
   { &search4_0.value, &search14_1.value },
   NULL,
};

   /* replace16_0_0 -> search4_0 in the cache */

/* replace16_0_1 -> search14_1 in the cache */
static const nir_search_expression replace16_0 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_ult16,
   { &search4_0.value, &search14_1.value },
   NULL,
};

/* replace16_1 -> search4_0 in the cache */

/* replace16_2 -> search14_1 in the cache */
static const nir_search_expression replace16 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b16csel,
   { &replace16_0.value, &search4_0.value, &search14_1.value },
   NULL,
};

   /* search17_0 -> search4_0 in the cache */

/* search17_1 -> search14_1 in the cache */
static const nir_search_expression search17 = {
   { nir_search_value_expression, 16 },
   false, false,
   0, 1,
   nir_op_umax,
   { &search4_0.value, &search14_1.value },
   NULL,
};

   /* replace17_0_0 -> search14_1 in the cache */

/* replace17_0_1 -> search4_0 in the cache */
static const nir_search_expression replace17_0 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_ult16,
   { &search14_1.value, &search4_0.value },
   NULL,
};

/* replace17_1 -> search4_0 in the cache */

/* replace17_2 -> search14_1 in the cache */
static const nir_search_expression replace17 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_b16csel,
   { &replace17_0.value, &search4_0.value, &search14_1.value },
   NULL,
};

   /* search18_0 -> search7_0 in the cache */

static const nir_search_variable search18_1 = {
   { nir_search_value_variable, 32 },
   1, /* b */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
};
static const nir_search_expression search18 = {
   { nir_search_value_expression, 32 },
   false, false,
   0, 1,
   nir_op_imin,
   { &search7_0.value, &search18_1.value },
   NULL,
};

   /* replace18_0_0 -> search7_0 in the cache */

/* replace18_0_1 -> search18_1 in the cache */
static const nir_search_expression replace18_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_ilt32,
   { &search7_0.value, &search18_1.value },
   NULL,
};

/* replace18_1 -> search7_0 in the cache */

/* replace18_2 -> search18_1 in the cache */
static const nir_search_expression replace18 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b32csel,
   { &replace18_0.value, &search7_0.value, &search18_1.value },
   NULL,
};

   /* search19_0 -> search7_0 in the cache */

/* search19_1 -> search18_1 in the cache */
static const nir_search_expression search19 = {
   { nir_search_value_expression, 32 },
   false, false,
   0, 1,
   nir_op_imax,
   { &search7_0.value, &search18_1.value },
   NULL,
};

   /* replace19_0_0 -> search18_1 in the cache */

/* replace19_0_1 -> search7_0 in the cache */
static const nir_search_expression replace19_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_ilt32,
   { &search18_1.value, &search7_0.value },
   NULL,
};

/* replace19_1 -> search7_0 in the cache */

/* replace19_2 -> search18_1 in the cache */
static const nir_search_expression replace19 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b32csel,
   { &replace19_0.value, &search7_0.value, &search18_1.value },
   NULL,
};

   /* search20_0 -> search7_0 in the cache */

/* search20_1 -> search18_1 in the cache */
static const nir_search_expression search20 = {
   { nir_search_value_expression, 32 },
   false, false,
   0, 1,
   nir_op_umin,
   { &search7_0.value, &search18_1.value },
   NULL,
};

   /* replace20_0_0 -> search7_0 in the cache */

/* replace20_0_1 -> search18_1 in the cache */
static const nir_search_expression replace20_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_ult32,
   { &search7_0.value, &search18_1.value },
   NULL,
};

/* replace20_1 -> search7_0 in the cache */

/* replace20_2 -> search18_1 in the cache */
static const nir_search_expression replace20 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b32csel,
   { &replace20_0.value, &search7_0.value, &search18_1.value },
   NULL,
};

   /* search21_0 -> search7_0 in the cache */

/* search21_1 -> search18_1 in the cache */
static const nir_search_expression search21 = {
   { nir_search_value_expression, 32 },
   false, false,
   0, 1,
   nir_op_umax,
   { &search7_0.value, &search18_1.value },
   NULL,
};

   /* replace21_0_0 -> search18_1 in the cache */

/* replace21_0_1 -> search7_0 in the cache */
static const nir_search_expression replace21_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_ult32,
   { &search18_1.value, &search7_0.value },
   NULL,
};

/* replace21_1 -> search7_0 in the cache */

/* replace21_2 -> search18_1 in the cache */
static const nir_search_expression replace21 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b32csel,
   { &replace21_0.value, &search7_0.value, &search18_1.value },
   NULL,
};

   /* search22_0 -> search4_0 in the cache */
static const nir_search_expression search22 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_fexp2,
   { &search4_0.value },
   NULL,
};

   /* replace22_0_0_0 -> search4_0 in the cache */
static const nir_search_expression replace22_0_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_f2f32,
   { &search4_0.value },
   NULL,
};
static const nir_search_expression replace22_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_fexp2,
   { &replace22_0_0.value },
   NULL,
};
static const nir_search_expression replace22 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_f2f16,
   { &replace22_0.value },
   NULL,
};

   /* search23_0 -> search4_0 in the cache */
static const nir_search_expression search23 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_flog2,
   { &search4_0.value },
   NULL,
};

   /* replace23_0_0_0 -> search4_0 in the cache */
/* replace23_0_0 -> replace22_0_0 in the cache */
static const nir_search_expression replace23_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_flog2,
   { &replace22_0_0.value },
   NULL,
};
static const nir_search_expression replace23 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_f2f16,
   { &replace23_0.value },
   NULL,
};

   /* search24_0 -> search4_0 in the cache */
static const nir_search_expression search24 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_fsin,
   { &search4_0.value },
   NULL,
};

   /* replace24_0_0_0 -> search4_0 in the cache */
/* replace24_0_0 -> replace22_0_0 in the cache */
static const nir_search_expression replace24_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_fsin,
   { &replace22_0_0.value },
   NULL,
};
static const nir_search_expression replace24 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_f2f16,
   { &replace24_0.value },
   NULL,
};

   /* search25_0 -> search4_0 in the cache */
static const nir_search_expression search25 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_fcos,
   { &search4_0.value },
   NULL,
};

   /* replace25_0_0_0 -> search4_0 in the cache */
/* replace25_0_0 -> replace22_0_0 in the cache */
static const nir_search_expression replace25_0 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_fcos,
   { &replace22_0_0.value },
   NULL,
};
static const nir_search_expression replace25 = {
   { nir_search_value_expression, 16 },
   false, false,
   -1, 0,
   nir_op_f2f16,
   { &replace25_0.value },
   NULL,
};

   /* search26_0 -> search0_0 in the cache */
static const nir_search_expression search26 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_f2b32,
   { &search0_0.value },
   NULL,
};

   /* replace26_0 -> search0_0 in the cache */

static const nir_search_constant replace26_1 = {
   { nir_search_value_constant, -1 },
   nir_type_float, { 0x0 /* 0.0 */ },
};
static const nir_search_expression replace26 = {
   { nir_search_value_expression, 32 },
   false, false,
   0, 1,
   nir_op_fneu32,
   { &search0_0.value, &replace26_1.value },
   NULL,
};

   /* search27_0 -> search0_0 in the cache */
static const nir_search_expression search27 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_i2b32,
   { &search0_0.value },
   NULL,
};

   /* replace27_0 -> search0_0 in the cache */

/* replace27_1 -> replace0_0 in the cache */
static const nir_search_expression replace27 = {
   { nir_search_value_expression, 32 },
   false, false,
   0, 1,
   nir_op_ine32,
   { &search0_0.value, &replace0_0.value },
   NULL,
};

   /* search28_0 -> search7_0 in the cache */
static const nir_search_expression search28 = {
   { nir_search_value_expression, 32 },
   false, false,
   -1, 0,
   nir_op_b2i32,
   { &search7_0.value },
   NULL,
};

   /* replace28_0 -> search7_0 in the cache */

static const nir_search_constant replace28_1 = {
   { nir_search_value_constant, 32 },
   nir_type_int, { 0x1 /* 1 */ },
};
static const nir_search_expression replace28 = {
   { nir_search_value_expression, 32 },
   false, false,
   0, 1,
   nir_op_iand,
   { &search7_0.value, &replace28_1.value },
   NULL,
};


static const struct transform bifrost_nir_lower_algebraic_late_state2_xforms[] = {
  { &search0, &replace0.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state3_xforms[] = {
  { &search1, &replace1.value, 0 },
  { &search2, &replace2.value, 0 },
  { &search3, &replace3.value, 0 },
  { &search4, &replace4.value, 0 },
  { &search5, &replace5.value, 0 },
  { &search6, &replace6.value, 0 },
  { &search7, &replace7.value, 0 },
  { &search8, &replace8.value, 0 },
  { &search9, &replace9.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state4_xforms[] = {
  { &search10, &replace10.value, 0 },
  { &search14, &replace14.value, 0 },
  { &search18, &replace18.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state5_xforms[] = {
  { &search11, &replace11.value, 0 },
  { &search15, &replace15.value, 0 },
  { &search19, &replace19.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state6_xforms[] = {
  { &search12, &replace12.value, 0 },
  { &search16, &replace16.value, 0 },
  { &search20, &replace20.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state7_xforms[] = {
  { &search13, &replace13.value, 0 },
  { &search17, &replace17.value, 0 },
  { &search21, &replace21.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state8_xforms[] = {
  { &search22, &replace22.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state9_xforms[] = {
  { &search23, &replace23.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state10_xforms[] = {
  { &search24, &replace24.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state11_xforms[] = {
  { &search25, &replace25.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state12_xforms[] = {
  { &search26, &replace26.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state13_xforms[] = {
  { &search27, &replace27.value, 0 },
};
static const struct transform bifrost_nir_lower_algebraic_late_state14_xforms[] = {
  { &search28, &replace28.value, 0 },
};

static const struct per_op_table bifrost_nir_lower_algebraic_late_table[nir_num_search_ops] = {
   [nir_op_ineg] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         2,
      },
   },
   [nir_search_op_b2f] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         3,
      },
   },
   [nir_op_imin] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         4,
      },
   },
   [nir_op_imax] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         5,
      },
   },
   [nir_op_umin] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         6,
      },
   },
   [nir_op_umax] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         7,
      },
   },
   [nir_op_fexp2] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         8,
      },
   },
   [nir_op_flog2] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         9,
      },
   },
   [nir_op_fsin] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         10,
      },
   },
   [nir_op_fcos] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         11,
      },
   },
   [nir_search_op_f2b] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         12,
      },
   },
   [nir_search_op_i2b] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         13,
      },
   },
   [nir_search_op_b2i] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         14,
      },
   },
};

const struct transform *bifrost_nir_lower_algebraic_late_transforms[] = {
   NULL,
   NULL,
   bifrost_nir_lower_algebraic_late_state2_xforms,
   bifrost_nir_lower_algebraic_late_state3_xforms,
   bifrost_nir_lower_algebraic_late_state4_xforms,
   bifrost_nir_lower_algebraic_late_state5_xforms,
   bifrost_nir_lower_algebraic_late_state6_xforms,
   bifrost_nir_lower_algebraic_late_state7_xforms,
   bifrost_nir_lower_algebraic_late_state8_xforms,
   bifrost_nir_lower_algebraic_late_state9_xforms,
   bifrost_nir_lower_algebraic_late_state10_xforms,
   bifrost_nir_lower_algebraic_late_state11_xforms,
   bifrost_nir_lower_algebraic_late_state12_xforms,
   bifrost_nir_lower_algebraic_late_state13_xforms,
   bifrost_nir_lower_algebraic_late_state14_xforms,
};

const uint16_t bifrost_nir_lower_algebraic_late_transform_counts[] = {
   0,
   0,
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state2_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state3_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state4_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state5_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state6_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state7_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state8_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state9_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state10_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state11_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state12_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state13_xforms),
   (uint16_t)ARRAY_SIZE(bifrost_nir_lower_algebraic_late_state14_xforms),
};

bool
bifrost_nir_lower_algebraic_late(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[1];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   condition_flags[0] = true;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        bifrost_nir_lower_algebraic_late_transforms,
                                        bifrost_nir_lower_algebraic_late_transform_counts,
                                        bifrost_nir_lower_algebraic_late_table);
      }
   }

   return progress;
}

